/* Copyright (C) 2007 Delta Networks Inc.
 * This module is used for DNI's routing functions, logged packets according to NETGEAR spec.
 * And we are the last match in the netfilter rules to reduce the number of rules.
 */
#include <stdio.h>
#include <netdb.h>
#include <string.h>
#include <stdlib.h>
#include <getopt.h>
#include <ctype.h>

#include <iptables.h>
#include <linux/netfilter_ipv4/ipt_log.h>


/* Function which prints out usage message. */
static void
help(void)
{
	printf("log match v%s options:\n"
		"--log-prefix string          add prefix before log information\n",
		IPTABLES_VERSION);
}


static struct option opts[] = {
	{ .name = "log-prefix",     .has_arg = 1, .flag = 0, .val = '1' },
	{ .name = 0 }
};


/* Initialize the match. */
static void
init(struct ipt_entry_match *m, unsigned int *nfcache)
{
	struct ipt_logs_info *loginfo = (struct ipt_logs_info *)m->data;
	loginfo->flags = LOG_SPORT;
}



static int
parse_string(const unsigned char *s, struct ipt_logs_info *info)
{
        if (strlen(s) <= PREFIX_LENGTH) strcpy(info->prefix, s);
        else {
		exit_error(PARAMETER_PROBLEM, "STRING too long `%s'", s);
		return 0;
	}
	return 1;
}

/* Function which parses command options; returns true if it
   ate an option */
static int
parse(int c, char **argv, int invert, unsigned int *flags,
      const struct ipt_entry *entry,
      unsigned int *nfcache,
      struct ipt_entry_match **match)
{
	struct ipt_logs_info *loginfo = (struct ipt_logs_info *)(*match)->data;

	switch (c) {
	case '1':
		check_inverse(optarg, &invert, &optind, 0);
		if (invert) {
			printf("Invert can't be used on --log-prefix\n");
			return 0;
		}
		if (! parse_string(argv[optind-1], loginfo))
			return 0;
		break;

	default:
		return 0;
	}
	return 1;
}


/* Prints out the matchinfo. */
static void
print(const struct ipt_ip *ip,
      const struct ipt_entry_match *match,
      int numeric)
{
	const struct ipt_logs_info *info =
	    (const struct ipt_logs_info*) match->data;

	if (info->prefix[0]) {
		printf("log-prefix: %s", info->prefix);
	}
}


/* Saves the union ipt_matchinfo in parseable form to stdout. */
static void
save(const struct ipt_ip *ip, const struct ipt_entry_match *match)
{
	const struct ipt_logs_info *info =
	    (const struct ipt_logs_info*) match->data;

	if (info->prefix[0]) {
		printf("prefix: %s", info->prefix);
	}
}


/* Final check */
static void
final_check(unsigned int flags)
{
        /* Nothing to do */
}


static struct iptables_match log = {
    .name          = "log",
    .version       = IPTABLES_VERSION,
    .size          = IPT_ALIGN(sizeof(struct ipt_logs_info)),
    .userspacesize = IPT_ALIGN(sizeof(struct ipt_logs_info)),
    .help          = &help,
    .init          = &init,
    .parse         = &parse,
    .final_check   = &final_check,
    .print         = &print,
    .save          = &save,
    .extra_opts    = opts
};


void _init(void)
{
	register_match(&log);
}
